package com.example.ueberwachungssystem.Detection;

import android.Manifest;
import android.annotation.SuppressLint;
import android.app.Activity;
import android.content.Context;
import android.content.pm.PackageManager;
import android.graphics.Bitmap;
import android.graphics.ImageFormat;
import android.media.Image;
import android.os.CountDownTimer;
import android.util.Log;
import android.view.Surface;
import android.widget.ImageView;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.camera.core.CameraSelector;
import androidx.camera.core.ExperimentalGetImage;
import androidx.camera.core.ImageAnalysis;
import androidx.camera.core.Preview;
import androidx.camera.core.VideoCapture;
import androidx.camera.lifecycle.ProcessCameraProvider;
import androidx.camera.view.PreviewView;
import androidx.core.app.ActivityCompat;
import androidx.core.content.ContextCompat;
import androidx.lifecycle.LifecycleOwner;

import com.google.common.util.concurrent.ListenableFuture;

import org.opencv.android.OpenCVLoader;
import org.opencv.android.Utils;
import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.imgproc.Imgproc;

import java.io.File;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ExecutionException;


/**
 * Video Detector inherits some methods from abstract Detector class (more info there)
 * USE FROM MAIN ACTIVITY:
 * VideoDetector vd = new VideoDetector(this);
 * */


@ExperimentalGetImage
public class VideoDetector extends Detector {

    // Calling Activity
    private final Context context;

    // Permission handling
    private static final int PERMISSION_REQUEST_CODE = 3691;

    // Camera Provider
    private ProcessCameraProvider cameraProvider;
    private final ImageAnalysis imageAnalysis;
    private final VideoCapture videoCapture;
    private final Preview preview;

    // Logic
    private boolean isDetecting = false;
    private boolean isRecording = false;
    private boolean allowReportViolation = false;

    // Image Processing
    private Mat previousImage = null;

    // Debugging
    private ImageView inputImageView = null;
    private ImageView outputImageView = null;


    // Recording
    private final String outputName = "video.mp4";


    // Parameters
    private static final float ALARM_THRESHOLD = 0.5f;   // Percent of pixels changed
    private static final long START_DELAY = 20000;   // milliseconds
    private static final android.util.Size IMAGE_RES = new android.util.Size(640, 480);


    private enum UseCase {
        ImageAnalysis,
        Preview,
        VideoCapture
    };


    /** Constructor */
    public VideoDetector(Context context) {
        super();
        this.context = context;
        this.imageAnalysis = setupImageAnalysis();
        this.videoCapture = setupVideoCapture();
        this.preview = new Preview.Builder().build();
    }

    /** Get State of the Detector */
    public boolean isRunning() {
        return isDetecting;
    }


    /** Starts the Video Detection */
    @Override
    public void startDetection() {
        // Check States
        if (isDetecting)
            return;
        // Return On Request Permissions
        if (!hasPermissions()) {
            getPermissions();
            return;
        }
        // Open CV startup check
        if (!OpenCVLoader.initDebug()) {
            Log.e("OpenCV", "Unable to load OpenCV!");
            return;
        } else
            Log.d("OpenCV", "OpenCV loaded Successfully!");
        // Get Process Camera Provider and start
        final ListenableFuture<ProcessCameraProvider> cameraProviderFuture = ProcessCameraProvider.getInstance(context);
        cameraProviderFuture.addListener(() -> {
            try {
                cameraProvider = cameraProviderFuture.get();
                isDetecting = true;
                bindCameraProvider(UseCase.ImageAnalysis);
            } catch (ExecutionException | InterruptedException e) {}
        }, ContextCompat.getMainExecutor(context));
    }

    /** Stops the Video Detection */
    @Override
    public void stopDetection() {
        if (!isDetecting || imageAnalysis == null)
            return;
        cameraProvider.unbind(imageAnalysis);
        cameraProvider.unbind(preview);
        isDetecting = false;
        allowReportViolation = false;
    }

    /** Permission handling */
    private boolean hasPermissions() {
        return ContextCompat.checkSelfPermission(context, Manifest.permission.CAMERA) == PackageManager.PERMISSION_GRANTED &&
                ContextCompat.checkSelfPermission(context, Manifest.permission.RECORD_AUDIO) == PackageManager.PERMISSION_GRANTED;
    }
    private void getPermissions() {
        if (!hasPermissions())
            ActivityCompat.requestPermissions((Activity) context, new String[]{android.Manifest.permission.CAMERA, Manifest.permission.RECORD_AUDIO}, PERMISSION_REQUEST_CODE);
    }



    /** Binds the Luminosity Analyzer (configure and run Analysis) */
    private void bindCameraProvider(UseCase useCase) {
        // Specify which Camera to use
        CameraSelector cameraSelector = new CameraSelector.Builder().requireLensFacing(CameraSelector.LENS_FACING_BACK).build();
        if(useCase == UseCase.ImageAnalysis && !cameraProvider.isBound(videoCapture)) {
            cameraProvider.unbindAll();
            cameraProvider.bindToLifecycle((LifecycleOwner) context, cameraSelector, imageAnalysis, preview);
            startViolationTimer();
        }
        if(useCase == UseCase.VideoCapture) {
            if(cameraProvider.isBound(imageAnalysis)) {
                cameraProvider.unbindAll();
                cameraProvider.bindToLifecycle((LifecycleOwner) context, cameraSelector, imageAnalysis, videoCapture);
            } else {
                cameraProvider.unbindAll();
                cameraProvider.bindToLifecycle((LifecycleOwner) context, cameraSelector, videoCapture);
            }
        }
    }


    /** Start delay until Violation Report is allowed */
    private void startViolationTimer() {
        new CountDownTimer((long) (START_DELAY), 100) {
            @Override
            public void onTick(long millisUntilFinished) {
            }
            @Override
            public void onFinish() {
                allowReportViolation = true;
            }
        }.start();
    }


    private ImageAnalysis setupImageAnalysis() {
        // Configure and create Image Analysis
        ImageAnalysis.Builder builder = new ImageAnalysis.Builder();
        builder.setTargetResolution(IMAGE_RES);
        builder.setBackpressureStrategy(ImageAnalysis.STRATEGY_KEEP_ONLY_LATEST);
        builder.setOutputImageFormat(ImageAnalysis.OUTPUT_IMAGE_FORMAT_YUV_420_888);
        ImageAnalysis imageAnalysis = builder.build();
        // Set Analyzer
        imageAnalysis.setAnalyzer(ContextCompat.getMainExecutor(context), imageProxy -> {
            if (imageProxy.getFormat() == ImageFormat.YUV_420_888) {
                Image image = imageProxy.getImage();
                assert image != null;

                // Violation Handling
                Mat processed = processImage(image);
                int n = OpenCVHelper.countNonZeroPixels(processed);
                int pixelCount = image.getWidth() * image.getHeight();
                float percentChanged = (float) n / pixelCount;

                // Violation Condition
                if (percentChanged * 100 > ALARM_THRESHOLD) {
                    if (allowReportViolation)
                        reportViolation("Video", n);
                }
            }
            imageProxy.close();
        });
        return imageAnalysis;
    }

    @SuppressLint("RestrictedApi")
    private VideoCapture setupVideoCapture() {
        return new VideoCapture.Builder()
                .setTargetRotation(Surface.ROTATION_0)
                .build();
    }

    @SuppressLint("RestrictedApi")
    public void startRecording() {
        // Check States
        if (isRecording){
            extendDetection();
            return;
        }

        // Return On Request Permissions
        if (!hasPermissions()) {
            getPermissions();
            return;
        }

        final ListenableFuture<ProcessCameraProvider> cameraProviderFuture = ProcessCameraProvider.getInstance(context);
        cameraProviderFuture.addListener(() -> {
            try {
                cameraProvider = cameraProviderFuture.get();
                isRecording = true;
                bindCameraProvider(UseCase.VideoCapture);

                File vidFile = new File(context.getFilesDir() + "/" + outputName);
                if (ActivityCompat.checkSelfPermission(context, Manifest.permission.RECORD_AUDIO) != PackageManager.PERMISSION_GRANTED) {
                    return;
                }
                videoCapture.startRecording(
                        new VideoCapture.OutputFileOptions.Builder(vidFile).build(),
                        context.getMainExecutor(),
                        new VideoCapture.OnVideoSavedCallback() {
                            @Override
                            public void onVideoSaved(@NonNull VideoCapture.OutputFileResults outputFileResults) {
                                Toast.makeText(context, "recording saved", Toast.LENGTH_SHORT).show();
                            }
                            @Override
                            public void onError(int videoCaptureError, @NonNull String message, @Nullable Throwable cause) {
                                Toast.makeText(context, "recording failed", Toast.LENGTH_SHORT).show();
                            }
                        }
                );
            } catch (ExecutionException | InterruptedException e) {}
        }, ContextCompat.getMainExecutor(context));
    }

    @SuppressLint("RestrictedApi")
    public void stopRecording(){
        videoCapture.stopRecording();
        cameraProvider.unbind(videoCapture);
        if (isDetecting)
            bindCameraProvider(UseCase.ImageAnalysis);
    }

    /** Process Image to be used for Motion Detection */
    private Mat processImage(Image image){
        // Image Transformation
        Mat imageMat = OpenCVHelper.extractYChannel(image);

        // Show Input Image
        if (inputImageView != null)
            OpenCVHelper.debugMat(imageMat, inputImageView);

        // Preprocess Image
        Mat preprocessed = imageMat;
        preprocessed = OpenCVHelper.addGaussianBlur(preprocessed, new Size(21, 21));
        preprocessed = OpenCVHelper.addBlur(preprocessed, new Size(3, 3));

        if (previousImage == null) {
            previousImage = preprocessed;
            return null;
        }

        // Process Image
        Mat processed = preprocessed.clone();
        processed = OpenCVHelper.thresholdPixels(processed, previousImage, 25);
        processed = OpenCVHelper.dilateBinaryMat(processed, new Size(3,3));
        processed = OpenCVHelper.dilateBinaryMat(processed, new Size(3,3));
        processed = OpenCVHelper.thresholdContourArea(processed, 500);

        // Output
        previousImage = preprocessed.clone();
        // Show Output Image
        if (outputImageView != null)
            OpenCVHelper.debugMat(processed, outputImageView);
        return processed;
    }

    public void debugProcessing(@NonNull ImageView inputImageView, @NonNull ImageView outputImageView){
        this.inputImageView = inputImageView;
        this.outputImageView = outputImageView;
    }

    public void setPreviewView(@NonNull PreviewView previewView) {
        // Create Preview
        if (this.preview != null)
            this.preview.setSurfaceProvider(previewView.getSurfaceProvider());
    }






    private static class OpenCVHelper{
        private OpenCVHelper() {}

        /** OpenCV helper methods **/
        private static Mat addGaussianBlur(Mat inputMat, Size kernelSize){
            Mat outputMat = new Mat();
            Imgproc.GaussianBlur(inputMat, outputMat, kernelSize, 0);
            return outputMat;
        }

        private static Mat addBlur(Mat inputMat, Size kernelSize){
            Mat outputMat = new Mat();
            Imgproc.blur(inputMat, outputMat, kernelSize);
            return outputMat;
        }

        private static Mat extractYChannel(@NonNull Image img) {
            ByteBuffer yBuffer = img.getPlanes()[0].getBuffer();
            byte[] yData = new byte[yBuffer.remaining()];
            yBuffer.get(yData);

            Mat yMat = new Mat(img.getHeight(), img.getWidth(), CvType.CV_8UC1);
            yMat.put(0, 0, yData);

            return yMat;
        }

        private static Mat thresholdPixels(Mat inputMat, Mat previousImage, int threshold){
            Mat diffImage = new Mat();
            Core.absdiff(inputMat, previousImage, diffImage);
            Mat binaryMat = new Mat();
            Imgproc.threshold(diffImage, binaryMat, threshold, 255, Imgproc.THRESH_BINARY);
            return binaryMat;
        }

        private static Mat thresholdContourArea(Mat inputMat, float areaThreshold){
            List<MatOfPoint> contours = new ArrayList<>();
            Mat hierarchy = new Mat();
            Imgproc.findContours(inputMat, contours, hierarchy, Imgproc.RETR_EXTERNAL, Imgproc.CHAIN_APPROX_SIMPLE);

            Mat outputMat = new Mat(inputMat.size(), inputMat.type(), new Scalar(0));
            // Iterate over the contours and draw only the larger contours on the outputMat
            for (MatOfPoint contour : contours) {
                double contourArea = Imgproc.contourArea(contour);
                if (contourArea > areaThreshold) {
                    Imgproc.drawContours(outputMat, Collections.singletonList(contour), 0, new Scalar(255), -1);
                }
            }
            // Apply the outputMat as a mask to the dilatedImage
            Mat maskedImage = new Mat();
            inputMat.copyTo(maskedImage, outputMat);
            return outputMat;
        }

        private static Mat dilateBinaryMat(Mat inputMat, Size kernelSize){
            Mat dilatedMat = new Mat();
            Mat kernel = Imgproc.getStructuringElement(Imgproc.MORPH_ELLIPSE, kernelSize);
            Imgproc.dilate(inputMat, dilatedMat, kernel);
            return dilatedMat;
        }

        private static int countNonZeroPixels(Mat inputImage) {
            if (inputImage != null)
                return Core.countNonZero(inputImage);
            else
                return 0;
        }


        private static void debugMat(Mat mat, ImageView imageView) {
            if (imageView == null || mat == null)
                return;

            Bitmap bitmap = Bitmap.createBitmap(mat.cols(), mat.rows(), Bitmap.Config.ARGB_8888);
            Utils.matToBitmap(mat, bitmap);

            // Display the bitmap in an ImageView
            imageView.setImageBitmap(bitmap);
        }
    }
}